/*
 * @(#)file      SnmpMibValidation.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.13
 * @(#)lastedit  04/05/10
 * 
 * Copyright 2003 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package validator;

import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Collection;
import java.util.List;
import java.util.ArrayList;

import com.sun.management.snmp.SnmpOidDatabaseSupport;

import com.sun.management.snmp.SnmpOidRecord;
import com.sun.management.snmp.SnmpVarBindList;
import com.sun.management.snmp.SnmpValue;
import com.sun.management.snmp.SnmpOid;
import com.sun.management.snmp.SnmpStatusException;
import com.sun.management.snmp.manager.SnmpRequestHandler;
import com.sun.management.snmp.manager.SnmpRequest;
import com.sun.management.snmp.manager.SnmpSession;
import com.sun.management.snmp.SnmpDefinitions;
import com.sun.management.snmp.SnmpVarBind;
import com.sun.management.snmp.SnmpNull;
import com.sun.management.snmp.SnmpString;
import com.sun.management.snmp.SnmpInt;
import com.sun.management.snmp.SnmpOpaque;
import com.sun.management.snmp.SnmpTimeticks;
import com.sun.management.snmp.SnmpGauge;
import com.sun.management.snmp.SnmpCounter;
import com.sun.management.snmp.SnmpCounter64;
import com.sun.management.snmp.SnmpIpAddress;
import com.sun.management.snmp.SnmpUnsignedInt;

import com.sun.management.snmp.manager.SnmpPeer;
import com.sun.management.snmp.manager.SnmpParams;

class ValidationThread extends Thread {
    private SnmpSession session;
    private SnmpMibValidation validation;
    private boolean error;
    
    ValidationThread(SnmpSession session,
		     SnmpMibValidation validation) {
	this.session = session;
	this.validation = validation;
    }

    public void run() {
	try {
	    validation.turboValidate(session); 
	}catch(Exception e) {
	    error = true;
	}
    }
    
    boolean errorOccured() {
	return error;
    }
}

class SnmpEntry {
    private SnmpVarBindList columns = new SnmpVarBindList();
    private SnmpOid index = null;
    private Map records = new HashMap();
    SnmpEntry(SnmpOid index) {
	this.index = index;
    }
    void addColumn(SnmpVarBind col, 
		   SnmpOidRecord record) throws SnmpStatusException {
	if(columns.contains(col))
	    throw new SnmpStatusException("Column already in entry.");
	columns.addVarBind(col);
	
	if(record != null)
	    records.put(col.getOid().toString(), record);
    }
    
    int nbColumns() {
	return columns.size();
    }
    
    Collection receivedIndexes() {
	return columns;
    }
    
    SnmpVarBindList getColumns() {
	return columns;
    }
    
    SnmpOidRecord getOidRecord(SnmpVarBind varbind) {
	return (SnmpOidRecord) records.get(varbind.getOid().toString());
    }
    
    SnmpOid index() {
	return index;
    }
}
    

/**
 * MIB validator. Validate MIBs.
 * 2 kind of validation :
 * - SnmpOidTable based MIB. For MIB processed using JDMK mibgen.
 * - Legacy MIB. Existing MIB for which no SnmpOidTable exists.
 * 
 * Some table specifics tests are done if <CODE> SnmpTableInfo </CODE> 
 * instances are provided.
 * 
 * The validator main algorithm is to walk the complete MIB then applies 
 * getNext requests to tables for wich some specific tests are provided.
 *
 * A validator instance can be used in multithreaded context. It allows you to
 * validate MIBs' thread-safety.
 * 
 * Associated logger is named com.sun.management.snmp.mibvalidator
 * Validation steps informations are logged in INFO level.
 * Debug traces are logged in FINEST level.
 * Error traces are logged in SEVERE level.
 */
public class SnmpMibValidation {
    
    /* Used in XML Parsing */
    private List oidtables = new ArrayList();
    private String name = "MIB Validation";
    private String description = "MIB Validation desc";
    private String type = "ALL";
    private int iteration = 1;
    private int thread = 1;
    /* Used in XML Parsing */

    private SnmpOidDatabaseSupport mibs = null;

    private List tables = null;

    private boolean init = false;

    private ArrayList oidsToGet = null;
    private HashMap tablesToGet = null;
    
    private SnmpPeer peer = null;
    
    private static ValidatorLogger logger = new ValidatorLogger(SnmpMibValidation.class);
    private SnmpOid start = null;
    private SnmpOid stop = null;
    private String snmprequest = "getnext";
    private SnmpOid getbulkStartOid = new SnmpOid("1.2");
    private int nbrepeaters = 5;
    
    private boolean doGetBulk = false;
    private boolean getAfterGetNext = false;
    
    /**
     * CALLED BY XML PARSER
     */
    SnmpMibValidation(String name,
		      String description,
		      String type,
		      int iteration,
		      int thread) {
	this((SnmpOid) null, (SnmpOid) null, (List) null);
	this.name = name;
	this.description = description;
	this.type = type;
	this.iteration = iteration;
	this.thread = thread;
	init = false;
    }
    
    /**
     * Create a validator with no arguments. the MIB is walked from 
     * oid 1.1 to oid 2.1. No specific tests is done on table entries.
     */
    public SnmpMibValidation() {
	this((SnmpOid) null, (SnmpOid) null, (List) null);
    }
    
    /**
     * The MIB is walked from 
     * oid 1.1 to oid 2.1. Provided specific tests are done on table entries.
     */
    public SnmpMibValidation(List tables) {
	this((SnmpOid) null, (SnmpOid) null, tables);
    }
    
    /**
     * Create a validator providing a set of oid tables. 
     * Some <CODE> SnmpTableInfo </DONE> can be provided. 
     * For each of them, tests will be applied to the mapping table.
     * @param mibs The set of oid records.
     * @param excludeList A list of oid logical names (<CODE>String</CODE>)
     * to exclude from the main set of oid records (non accessible oids).
     * @param tables An array of table desciptions.
     */
    public SnmpMibValidation(SnmpOidDatabaseSupport mibs,
			     List excludeList,
			     List tables) {
	this.tables = tables;

	makeDataStructures(excludeList, mibs);
	
	init = true;
    }

    /**
     * Create a validator providing a walk until start and stop oid. 
     * Some <CODE> SnmpTableInfo </DONE> can be provided. 
     * For each of them, tests will be applied to the mapping table.
     * @param start The oid from which a snmpwalk will be started.
     * @param stop The oid from which a snmpwalk will be stopped.
     * @param tables An array of table desciptions.
     */
    public SnmpMibValidation(SnmpOid start,
			     SnmpOid stop,
			     List tables) {
	this.start = start;
	this.stop = stop;
	this.tables = tables;
	init = true;
    }

    /**
     * Start the MIB validation process.
     * @param session A session used to make snmp requests.
     * @throws SnmpStatusException If any error occurs.
     */
    public void validate(SnmpSession session) throws SnmpStatusException {
	init();
	boolean error = false;
	System.out.println("Validation [" + name + "] nb iterations [" + iteration +"], nb threads ["+ thread + "], type [" + type + "]\n" +
			   "agent [" + peer.getDestAddr()+ 
			   ":" + peer.getDestPort() +"]");
	for(int i = 0; i < iteration; i++) {
	    System.out.println("\titeration [" + (i+1) +"] started");
	    try {
		if(thread == 1) {
		    turboValidate(session);	    	
		    System.out.println("\titeration [" + (i+1) +"] done");
		    continue;
		}
		ValidationThread[] threads = new ValidationThread[thread];
		for(int j = 0; j < thread; j++) {
		    threads[j] = new ValidationThread(session, this);
		    logger.debug("", "Thread-" + (j+1) +" start");
		    threads[j].setName("Thread-"+(j+1));
		    threads[j].start();
		}
		for(int j = 0; j < thread; j++) {
		    threads[j].join();
		    if(threads[j].errorOccured()) {
			logger.error("", "Error in thread [" + (j+1) +"]");
			throw new Exception("Error occured");
		    }
		    logger.debug("", "Thread [" + (j+1) +"] end");
		}
	    }catch(Exception e) {
		//e.printStackTrace();
		error = true;
	    }
	    System.out.println("\titeration [" + (i+1) +"] done");
	}
	
	if(error)
	    throw new SnmpStatusException("FAILED. " +
					  "Check SEVERE traces for detail");
	
	System.out.println("Validation [" + name + "]. SUCCESS \n\n");
    }

    /**
     * First time validate is called, data structures are computed
     */
    synchronized void init() {
	if(init) return;
	if(oidtables.size() != 0) {
	    //First merge ALL oid tables and exclude lists.
	    SnmpOidDatabaseSupport mibs = new SnmpOidDatabaseSupport();
	    ArrayList excludeList = new ArrayList();
	    Iterator it = oidtables.iterator();
	    while(it.hasNext()) {
		SnmpMibValidationOidTable table = (SnmpMibValidationOidTable) it.next();
		mibs.add(table.getOidTable());
		List exclude = table.getExcludeList();
		if(exclude == null) continue;
		Iterator i1 = exclude.iterator();
		while(i1.hasNext()) {
		    excludeList.add((String) i1.next());
		}	
	    }
	    makeDataStructures(excludeList, mibs);
	}
	
	init = true;
    }
    
    private void makeDataStructures(List excludeList, SnmpOidDatabaseSupport mibs) {
	if(mibs != null) {
	    
	    this.mibs = mibs;
	    
	    //Search for ALL tables.
	    oidsToGet = new ArrayList();
	    tablesToGet = new HashMap();
	    Iterator it = mibs.getAllEntries().iterator();
	    while(it.hasNext()) {
		SnmpOidRecord record = (SnmpOidRecord) it.next();
		if(record.getType().equals("TA")) {
		    tablesToGet.put(record.getOid(),
				    new SnmpTable(record.getOid(),
						  record.getName(),
						  getDescription(record.getOid())));
		}
	    }
	    
	    it = mibs.getAllEntries().iterator();
	    
	    // Then all oid
	    while(it.hasNext()) {
		SnmpOidRecord record = (SnmpOidRecord) it.next();
		String name = record.getName();
	
		if(!record.getType().equals("EN") &&
		   !record.getType().equals("TA") &&
		   !record.getType().equals("NT") &&
		   !record.getType().equals("ID") ){
		    
		    if( (excludeList == null) || 
			(excludeList != null && 
			 !excludeList.contains(name))){
			SnmpTable table = getMatchingTable(record.getOid());
			
			if(table == null)
			    oidsToGet.add(record);
			else
			    table.addColumn(record);
		    }
		}
	    }
	}
    }

    /**
     * Start the MIB tables ONLY validation process. 
     * @param session A bound session to a running agent.
     * @throws SnmpStatusException If any error occurs.
     */
    public void validateTables(SnmpSession session) 
	throws SnmpStatusException {
	logger.debug("["+Thread.currentThread().getName() + "]",
		    "STARTING TABLES validation.");
	if(!isMetaProvided()) {
	    logger.debug("", "Start Tables validation. " + 
			 "No oid table provided.");
	    validateTablesWithoutOidTable(session);
	} else {
	    logger.debug("", "Start Tables validation. " + 
			 "Oid table provided.");
	    validateTablesWithOidTable(session);
	}

	logger.debug("["+Thread.currentThread().getName() + "]",
		    "TABLES validation OK.");
    }
    
    /**
     * Start the MIB validation process. Tables tests are not processed.
     * @param session A bound session to a running agent.
     * @throws SnmpStatusException If any error occurs.
     */
    public void validateMib(SnmpSession session) throws SnmpStatusException {
	if(!isMetaProvided()) {
	    logger.debug("", "Start MIB validation. No oid table.");
	    validateMibWithoutOidTable(session);
	} else {
	    logger.debug("", "Start MIB validation. Oid table.");
	    validateMibWithOidTable(session);
	}
    }

    void turboValidate(SnmpSession session) throws SnmpStatusException {
	boolean error = false;
	
	try {
	    if(type == null || type.equals("ALL") || type.equals("UNARY"))
		validateMib(session);
	}catch(Exception e) {
	    error = true;
	}
	
	try {
	    if(type == null || type.equals("ALL") || type.equals("SCALARE"))
		validateTables(session);
	}catch(Exception e) {
	    error = true;
	}
	
	if(error) {
	    logger.error("", "validation [" + name + "]. FAILED\n\n");
	    throw new SnmpStatusException("validation [" + name + "]. FAILED");
	}
    }

    private void validateMibWithOidTable(SnmpSession session) 
	throws SnmpStatusException {
	Iterator it = oidsToGet.iterator();
	boolean error = false;
	int counter = 0;

	//test GetBulk requests
	if(isGetBulkToValidate()) {
	    if(oidsToGet.size() != 0) {
		activateGetBulk(session);
	    }
	}
	
	while(it.hasNext()) {
	    SnmpOidRecord record = (SnmpOidRecord) it.next();
	    String name = record.getName();
	    String o = record.getOid();
	    SnmpTypeChecker type = (SnmpTypeChecker) SnmpTypeChecker.checkers.get(record.getType());
	    if(type == null) {
		logger.error(name, 
			     "Invalid type [" + record.getType() + "]");
		error = true;
		continue;
	    }
	    try {
		SnmpOid oid = new SnmpOid(o + ".0");
		SnmpValue value = get(session, oid);
		logger.debug(name,
			     "Get oid[" + oid +"] value [" + value + "].");
		if(!type.isValidSnmpValue(value)) {
		    logger.error(name,
				 "Snmp error [" + oid + 
				 "]. Invalid SnmpValue. Expected [" + 
				 type.getValueClass() + "] received [" +
				 value.getClass() +"]");
		    error = true;
		}
	    }catch(SnmpStatusException e) {
		logger.error(name,
			     "Snmp error, error [" +  
			     SnmpRequest.snmpErrorToString(e.getStatus()) +
			     " ] index [" + e.getErrorIndex() + "]");
		error = true;
	    }
	    counter += 1;
	}	

	if(error) {
	    throw new SnmpStatusException("Error while validating MIB" + 
					  "Check SEVERE traces for details.");
	}
	System.out.println("\tthread ["+ Thread.currentThread().getName() + 
			   "]" + " validation OK. Validated [" + 
			   counter + "] oids.");
    }
    
    private void validateMibWithoutOidTable(SnmpSession session) 
	throws SnmpStatusException {

	if(start == null)
	    start = new SnmpOid("1.1");
	if(stop == null)
	    stop = new SnmpOid("2.1");
	SnmpVarBindList list = new SnmpVarBindList();
	SnmpVarBind var = new SnmpVarBind(start);
	list.add(var);
	
	if(isGetBulkToValidate()) {
	    activateGetBulk(session);
	}
	
	if(snmprequest.equals("getnext"))
	    doitWithGetNext(session, list);
	else
	    if(snmprequest.equals("walkuntil"))
		doitWithWalkUntil(session, list);
	    else
		System.out.println("["+Thread.currentThread().getName() + "]"+
				   "REQUEST [" + snmprequest +
				   "] NOT YET IMPLEMENTED");
    }
    
    private void doitWithGetNext(SnmpSession session,
				 SnmpVarBindList list) 
	throws SnmpStatusException {
	int counter = 0, totalCounter = 0;
	while(true) {
	    SnmpVarBindList result = getNext(session, list);
	    SnmpVarBind avar = result.getVarBindAt(0);
	    if(avar.hasVarBindException() || 
	       (avar.getOid().compareTo(stop)>= 0)) {
		logger.debug("", "MIB coverage completed  [" + avar + "]");
		break;
	    }    
	    list = result;
	    
	    if(isGetAfterGetNext()) {
		//Testing get request on the received varbind list
		get(session, result);
	    }
	    
	    counter ++;
	    
	    if(counter >= 100) {
		System.out.println("\tTotal received [" + 
			    totalCounter +"] for [" + name + "]");
		counter = 0;
	    }
	    
	    totalCounter +=1;
	}
	System.out.println("\tthread ["+ Thread.currentThread().getName() + 
			   "]" + " validation OK. Validated [" + 
			   totalCounter + "] oids.");
    }

    private void doitWithWalkUntil(SnmpSession session, 
				   SnmpVarBindList list) 
	throws SnmpStatusException {
	
	WalkUntilListener listener = 
	    new WalkUntilListener(Thread.currentThread().getName(),
				  session);
	SnmpRequest request = null;
	if(peer == null)
	    request = session.snmpWalkUntilRequest(listener, 
						   list,
						   stop);
	else
	    request = session.snmpWalkUntilRequest(peer,
						   listener, 
						   list,
						   stop);
	listener.waitDone();
	if(listener.exceptionOccured()) {
	    SnmpStatusException e = listener.getException();
	    logger.error("validate",
			 "Snmp error, error [" +  
			 SnmpRequest.snmpErrorToString(e.getStatus()) +
			 " ] index [" + e.getErrorIndex() +
			 "]");
	    throw e;
	}
	
	System.out.println("\tthread ["+ Thread.currentThread().getName() + 
			   "]" + " validation OK. Validated [" + 
			    listener.getNbReceivedResponses() + "] oids.");
    }
  
    private void validateTablesWithOidTable(SnmpSession session) 
	throws SnmpStatusException {
	Iterator it = tablesToGet.values().iterator();
	while(it.hasNext()) {
	    SnmpTable t = (SnmpTable) it.next();
	    SnmpTableInfo table = t.getDescription();
	    String name = table.getDescription();
	    String label = (name == null ? t.getName() : name);
	    System.out.println("["+Thread.currentThread().getName() + "]" + 
			       " Validating table [" + 
			       label +"]");
	    SnmpVarBindList vblst = t.getColumns();

	    Map result = getTableEntries(session, 
					 label,
					 table.getOid(),
					 vblst);
	    //Checking field Types
	    
	    checkTable(result, table, label);
	}
    }
    
    private void validateTablesWithoutOidTable(SnmpSession session) 
	throws SnmpStatusException {
	if(tables == null) return;
	Iterator it = tables.iterator();
	while(it.hasNext()) {
	    SnmpTableInfo table = (SnmpTableInfo) it.next();
	    String name = table.getDescription();
	    String label = (name == null ? table.getOid().toString() : name);
	    System.out.println("["+Thread.currentThread().getName() + "]"+ 
			       " Validating table [" + label +"]");
	    SnmpVarBindList vblst = discoverColumns(session, table.getOid());
	    Map result = getTableEntries(session,
					 label,
					 table.getOid(),
					 vblst);

	    checkTable(result, table, label);
	}
    }
    
    private void activateGetBulk(SnmpSession session) 
	throws SnmpStatusException {
	if(peer != null) {
	    SnmpParams p = peer.getParams();
	    if(p.getProtocolVersion() == SnmpDefinitions.snmpVersionOne) {
		logger.debug("","SNMP Version 1 doesn't support GetBulk");
		return;
	    }
	}
	System.out.println("\tthread ["+ Thread.currentThread().getName() + 
			   " getbulk from [" + getbulkStartOid + 
			   "] repeaters [" +
			   nbrepeaters + "]");
	
	// Sending a Getbulk(0,n,start) and see that it is equal to n getnext
	// starting at start
	SnmpVarBindList lst = new SnmpVarBindList();
	lst.addVarBind(new SnmpVarBind(getbulkStartOid));
	SnmpVarBindList b = getBulk(session, 
				    0, 
				    nbrepeaters, 
				    lst);
	List bulk = new ArrayList();
	Iterator it = b.iterator();
	while(it.hasNext()) {
	    SnmpVarBind var = (SnmpVarBind) it.next();
	    bulk.add(var.getOid().toString());
	}
	
	List next = new ArrayList();
	SnmpOid prev = getbulkStartOid;
	for(int i = 0; i < nbrepeaters; i++) {
	    prev = getNext(session, prev);
	    next.add(prev.toString());
	}
	
	if(!bulk.equals(next)) {	    
	    logger.debug("", "GetBulk List : " + bulk);
	    logger.debug("", "GetNext List : " + next);
	    throw new SnmpStatusException("GetNext and GetBulk doesn't return"+
					  " the same oids");
		
	}
	logger.debug("["+Thread.currentThread().getName() + "]",
		    "Getbulk validation OK. Received [" + next.size() +
		    "] oid");
    }

    private void checkTable(Map result, 
			    SnmpTableInfo table,
			    String label) throws SnmpStatusException {
	boolean error = false;
	final int size = result.values().size();
	
	String name=table.getDescription();
       
	//First test emptyness.
	if(size == 0 && !table.isEmptyAllowed()) {
	    logger.error("", "Table [" + label +"]\n" +
			 "Entries MUST be present.");
	    error = true;
	}
	
	//Then nb entries.
	final int expectedNbEntries = table.getExpectedNbEntries();
	if((expectedNbEntries != -1) &&
	   (size != expectedNbEntries) ){
	    logger.error("", "Table [" + label + "]\n" +
			 "Nb entries [" + size +
			 "] doesn't match expected one [" + 
			 expectedNbEntries +"]");
	    error = true;
	}
	
	//Then expected indexes
	List expectedEntries = table.getExpectedIndexes();
	if(expectedEntries != null) {
	    Collection keys = result.keySet();
	    if(!keys.containsAll(expectedEntries)) {
		logger.error("", "Table[" + label
			     + "]\nExpected entries mismatch");
		Iterator it = expectedEntries.iterator();
		while(it.hasNext()) {
		    String oid = (String) it.next();
		    if(!keys.contains(oid)) {
			logger.debug(name, "Expected entry [" + oid +
				     "] not found");
		    }
		}
		error = true;
	    }
	    
	    //Then strictness of expected indexes
	    if(table.isStrictIndexes()) {
		if(size != expectedEntries.size()) {
		    logger.error("", "Table [" + label + 
				 "]\nStrict indexes failed. Nb indexes [" +
				 size + "] expected [" + 
				 expectedEntries.size() + "]");
		    error = true;
		}
	    }
	}
	
	//Then nb columns
	final int nbcolumns = table.getNbColumns();
	if(nbcolumns != -1) {
	    Iterator values = result.values().iterator();
	    while(values.hasNext()) {
		SnmpEntry entry = (SnmpEntry) values.next();
		if(entry.nbColumns() != table.getNbColumns()) {
		    logger.error("", "Table [" + label + 
				 "]\n Number of columns [" + 
				 entry.nbColumns() + "] doesnt match " + 
				 "expected [" + nbcolumns + "] for entry [" +
				 entry.index() +"]");
		    Collection indexes = entry.receivedIndexes();
		    Iterator iter = indexes.iterator();
		    while(iter.hasNext()) {
			logger.trace(name, "Received column [ " + 
				     iter.next() +
				     "]");
		    }
		    error = true;
		}
		SnmpVarBindList columns = entry.getColumns();
		Iterator vars = columns.iterator();
		while(vars.hasNext()) {
		    SnmpVarBind var = (SnmpVarBind) vars.next();
		    SnmpOidRecord record = entry.getOidRecord(var);
		    if(record != null) {
			SnmpTypeChecker type = (SnmpTypeChecker) SnmpTypeChecker.checkers.get(record.getType());
			if(type == null) {
			    logger.error(name, 
					 "Invalid type [" + record.getType() + "]");
			    error = true;
			    continue;
			}
			if(!type.isValidSnmpValue(var.getSnmpValue())) {
			    logger.error(name,
					 "Snmp error [" + var.getOid() + 
					 "]. Invalid SnmpValue. Expected [" + 
					 type.getValueClass() + 
					 "] received [" +
					 var.getSnmpValue().getClass() +"]");
			    error = true;
			}
		    }
		}
	    }
	}
	
	if(error) {
	    throw new SnmpStatusException("Errors occured for table [" +
					  label +"]");
	}
    }

    private SnmpVarBindList discoverColumns(SnmpSession session,
					    SnmpOid oid) 
	throws SnmpStatusException {
	SnmpOid previousOid = oid;
	SnmpVarBindList list = new SnmpVarBindList();
	while(true) {
	    SnmpOid col = getNext(session, previousOid);
	    if(col.toString().startsWith(oid.toString())) {
		long[] entryOid = col.longValue(false);
		long[] tableOid = oid.longValue(false);
		long[] columnOid = new long[tableOid.length + 2];
		for(int i = 0; i < columnOid.length; i++) {
		    columnOid[i] = entryOid[i];
		}
		
		SnmpOid column = new SnmpOid(columnOid);
		logger.debug("discoverColumns", "New column : [" + 
			     column +"]");
		list.addVarBind(new SnmpVarBind(column));
		//Compute the next 
		columnOid[columnOid.length -1] += 1;
		previousOid = new SnmpOid(columnOid);
	    } else
		break;
	}
	return list;
    }
    
    private Map getTableEntries(SnmpSession session,
				String tableName,
				SnmpOid oid,
				SnmpVarBindList previousEntries) 
	throws SnmpStatusException {
	HashMap map = new HashMap();
	while(true) {
	    logger.debug("GetTableEntries", "table: " + tableName);
	    logger.debug("GetNext", "Getting nextof oid [" + 
			 previousEntries +"]");
	    SnmpVarBindList result = getNext(session, previousEntries);
	    //Test if still in the table
	    if(result.size() > 0 && result.getVarBindAt(0) != null) {
		SnmpOid columnOid = result.getVarBindAt(0).getOid();
		SnmpOid previousColumnOid = 
		    previousEntries.getVarBindAt(0).getOid();
		if(!switched(oid,tableName,columnOid, previousColumnOid) && 
		   columnOid.toString().startsWith(oid.toString())) {
		    
		    if(isGetAfterGetNext()) {
			//Testing get request on the received varbind list
			get(session, result);
		    }
		    
		    SnmpOid index = extractIndex(columnOid, oid);
		    
		    SnmpEntry entry = (SnmpEntry) map.get(index.toString());
		    
		    if(entry == null) {
			logger.debug(tableName, "Creating entry oid [" + 
				     index +"]");
			entry = new SnmpEntry(index);
			map.put(index.toString(), entry);
		    }
		    
		    Iterator it = result.iterator();
		    while(it.hasNext()) {
			SnmpVarBind var = (SnmpVarBind) it.next();
			SnmpOid fieldOid = extractFieldOid(var.getOid(), 
							   oid);
			SnmpOidRecord record = null;
			if(isMetaProvided())
			    record = mibs.resolveVarOid(fieldOid.toString());
			
			entry.addColumn(var, record);
		    }
		    previousEntries = entry.getColumns();
		} else
		    break;
	    } else
		break;
	}
	logger.debug("["+Thread.currentThread().getName() + "]", "Number of entries [" + map.size() + "]");
	return map;
    }
    
    private boolean switched(SnmpOid table,
			     String tableName,
			     SnmpOid next, 
			     SnmpOid previous) {
	
	long[] t = table.longValue(false);
	long[] n = next.longValue(false);
	long[] p = previous.longValue(false);
	boolean result = false;
	final int fieldCount = t.length + 2;
	int i = 0;
	
	while ((i < fieldCount) && (n[i] == p[i]))
	    i++ ;
	
	result = (i == fieldCount);
	
	if(!result)
	    logger.debug("switched", 
			 "End of table [" + tableName + "] reached.");
	
	return !result;
    }
    
    private SnmpOid extractFieldOid(SnmpOid oid, SnmpOid table) {
	long[] columnOid = oid.longValue(false);
	long[] tableOid = table.longValue(false);
	int size = tableOid.length + 2;
	long[] field = new long[size];
	for(int i = 0; i < size; i++) {
	    field[i] = columnOid[i];
	}
	return new SnmpOid(field);
    }

    private SnmpOid extractIndex(SnmpOid oid, SnmpOid table) {
	long[] columnOid = oid.longValue(false);
	long[] tableOid = table.longValue(false);
	int size = columnOid.length - tableOid.length - 2;
	long[] index = new long[size];
	for(int i = tableOid.length + 2, j = 0; j < index.length; i++, j++) {
	    index[j] = columnOid[i];
	}
	return new SnmpOid(index);
    }
    
    private SnmpVarBindList getBulk(SnmpSession session, 
				    int nonrepeaters, 
				    int repeaters, 
				    SnmpVarBindList vblst)
	throws SnmpStatusException {
	SnmpRequest request = null;
	if(peer == null) 
	    request = session.snmpGetBulkRequest(null,
						 vblst,
						 nonrepeaters,
						 repeaters);
	else
	    request = session.snmpGetBulkRequest(peer,
						 null,
						 vblst,
						 nonrepeaters,
						 repeaters);
	
	final boolean completed = request.waitForCompletion(10000);
	
	// Check for a timeout of the request.
	//
	if (completed == false) {
	    logger.error("", "GetBulk Request timed out." +
			 " Check reachability of agent");
	    throw new SnmpStatusException("GetBulk Request timed out. " + 
					  "Check reachability of agent");
	}
	
	// Now we have a response. Check if the response contains 
	// an error.
	//
	final int errorStatus = request.getErrorStatus();
	
	if (errorStatus != SnmpDefinitions.snmpRspNoError) {
	    logger.error("",
			 "Snmp error, error [" +  
			 SnmpRequest.snmpErrorToString(errorStatus) +
			 " ] index [" + request.getErrorIndex() + "]");
	    
	    throw new SnmpStatusException("GetNext request. error [" +  
					  SnmpRequest.snmpErrorToString(errorStatus) +
					  " ] index [" + 
					  request.getErrorIndex() + "]");
	}
	
	final SnmpVarBindList result = request.getResponseVarBindList();
	
	return result;
    }
    
    private SnmpOid getNext(SnmpSession session,
			    SnmpOid oid) throws SnmpStatusException  {
	SnmpVarBindList l = new SnmpVarBindList();
	SnmpVarBind var = new SnmpVarBind(oid);
	l.addVarBind(var);
	SnmpVarBindList resp = getNext(session, l);
	
	return resp.getVarBindAt(0).getOid();
    }

    private SnmpVarBindList getNext(SnmpSession session, 
				    SnmpVarBindList vblst) 
	throws SnmpStatusException {
	
	SnmpRequest request = null;
	if(peer == null) 
	    request = session.snmpGetNextRequest(null,
						 vblst);
	else
	    request = session.snmpGetNextRequest(peer,
						 null,
						 vblst);
	
	final boolean completed = request.waitForCompletion(10000);
	
	// Check for a timeout of the request.
	//
	if (completed == false) {
	    logger.error("", "GetNext Request timed out." +
			 " Check reachability of agent");
	    throw new SnmpStatusException("GetNext Request timed out. " + 
					  "Check reachability of agent");
	}
	
	// Now we have a response. Check if the response contains 
	// an error.
	//
	final int errorStatus = request.getErrorStatus();
	
	if (errorStatus != SnmpDefinitions.snmpRspNoError) {
	    logger.error("",
			 "Snmp error, error [" +  
			 SnmpRequest.snmpErrorToString(errorStatus) +
			 " ] index [" + request.getErrorIndex() + "]");
	    
	    throw new SnmpStatusException("GetNext request. error [" +  
					  SnmpRequest.snmpErrorToString(errorStatus) +
					  " ] index [" + 
					  request.getErrorIndex() + "]");
	}
	
	final SnmpVarBindList result = request.getResponseVarBindList();
	
	return result;
    }

    private SnmpVarBindList get(SnmpSession session, 
				SnmpVarBindList list) throws SnmpStatusException {
	SnmpRequest request = null;
	if(peer == null) 
	    request = session.snmpGetRequest(null,
					     list);
	else
	    request = session.snmpGetRequest(peer,
					     null,
					     list);
	
	final boolean completed = request.waitForCompletion(10000);
	
	// Check for a timeout of the request.
	//
	if (completed == false) {
	    logger.error("", "Get Request timed out." +
			 " Check reachability of agent");
	    throw new SnmpStatusException("Get Request timed out. " + 
					  "Check reachability of agent");
	}
	
	// Now we have a response. Check if the response contains 
	// an error.
	//
	final int errorStatus = request.getErrorStatus();
	
	if (errorStatus != SnmpDefinitions.snmpRspNoError) {
	    logger.error("",
			 "Snmp error, error [" +  
			 SnmpRequest.snmpErrorToString(errorStatus) +
			 " ] index [" + request.getErrorIndex() + "]");
	    
	    throw new SnmpStatusException("Get request. error [" +  
					  SnmpRequest.snmpErrorToString(errorStatus) +
					  " ] index [" + 
					  request.getErrorIndex() + "]");
	}
	
	final SnmpVarBindList result = request.getResponseVarBindList();
	
	return result;
    }

    private SnmpValue get(SnmpSession session, 
			  SnmpOid oid) 
	throws SnmpStatusException {
	SnmpVarBindList list = new SnmpVarBindList();
	list.addVarBind(new SnmpVarBind(oid));
	
	final SnmpVarBindList result = get(session, list);
	
	return result.getVarBindAt(0).getSnmpValue();
    }
    
    private SnmpTableInfo getDescription(String oid) {
	
	if(tables == null)
	    return null;
	
	Iterator it = tables.iterator();
	while(it.hasNext()) {
	    SnmpTableInfo table = (SnmpTableInfo) it.next();
	    if(table.getOid().toString().equals(oid))
		return table;
	}
	
	return null;
    }
    
    private boolean isMetaProvided() {
	return oidsToGet != null;
    }

    private boolean isGetBulkToValidate() {
	return doGetBulk;
    }
    
    private boolean isGetAfterGetNext() {
	return getAfterGetNext;
    }
    
    private SnmpTable getMatchingTable(String oid) {
	Iterator it = tablesToGet.values().iterator();
	while(it.hasNext()) {
	    SnmpTable t = (SnmpTable) it.next();
	    if(oid.startsWith(t.getOid().toString()))
		return t;
	}
	return null;
    }
    
    /* METHODS CALLED DURING XML parsing */
    
    void addOidTable(SnmpMibValidationOidTable table) {

	if(init) return;

	oidtables.add(table);
    }
    
    void setTableInfos(List tables) {
	if(init) return;
	
	this.tables = tables;
    }
    
    void addTreeLimits(SnmpOid start,
		       SnmpOid stop,
		       String snmprequest) {
	if(init) return;
		
	this.start = start;
	this.stop = stop;
	this.snmprequest = snmprequest;
    }

    void setPeer(SnmpPeer agent) {
	if(init) return;

	this.peer = agent;
    }

    void setGetBulkRequest(SnmpOid start,
			   int nbrepeaters) {
	this.getbulkStartOid = start;
	this.nbrepeaters = nbrepeaters;
	doGetBulk = true;
    }

    void setGetAfterGetNext(boolean b) {
	getAfterGetNext = b;
    }

    /* PRIVATE CLASSES */
    
    private class WalkUntilListener  implements SnmpRequestHandler {
	private static final int RUNNING = 1;
	private static final int DONE = 2;
	private SnmpStatusException e = null;
	private int status  = RUNNING;
	private int counter = 0;
	private int totalCounter = 0;
	private String name = null;
	private SnmpSession session = null;

	WalkUntilListener(String name, SnmpSession session) {
	    this.name = name;
	    this.session = session;
	}

	public boolean exceptionOccured() {
	    return (e != null);
	}

	public int getNbReceivedResponses() {
	    return totalCounter;
	}

	public SnmpStatusException getException() {
	    return e;
	}

	public void waitDone() {
	    try {
		synchronized(this) {
		    if(status == DONE)
			return;
		    else
			super.wait();
		}
	    }
	    catch(InterruptedException x) {
		
	    }
	}

	public void processSnmpInternalError(SnmpRequest request, 
					     String errmsg) {
	    logger.error("", errmsg);
	}
	
	public void processSnmpPollTimeout(SnmpRequest request) {
	    logger.error("", 
			 "Timeout, is the agent running?");
	}
	
	public void processSnmpPollData(SnmpRequest request, 
					int errStatus, 
					int errIndex, 
					SnmpVarBindList vblist) {
	    
	    final int errorStatus = request.getErrorStatus();
	    
	    if(errorStatus == SnmpDefinitions.snmpRspEndOfTable) {
		synchronized(this) {
		    status = DONE;
		    notify();
		}
		return;
	    }
	    
	    if (errorStatus != SnmpDefinitions.snmpRspNoError) {
		e = new SnmpStatusException(request.getErrorStatus(),
					    request.getErrorIndex());
		synchronized(this) {
		    status = DONE;
		    notify();
		}
	    }
	    
	    int size = request.getResponseVarBindList().getVarBindCount();
	    
	    for (int i = 0; i < size; i++) {
		SnmpVarBind b = 
		    request.getResponseVarBindList().getVarBindAt(i);
		
		String oid = b.getOid().toString();
		SnmpValue value = b.getSnmpValue();
		logger.debug("","Response oid ["+
			     oid +"] value [" + value +"]");
		if(totalCounter == 0 && counter == 0) {
		    logger.debug("", 
				 "Validating MIB [" +
				 oid +"](first oid)");
		}
		counter +=1;
	    }
	    
	    if(isGetAfterGetNext()) {
		try {
		    //Testing get request on the received varbind list
		    SnmpVarBindList l = 
			get(session, request.getResponseVarBindList());
		}catch(SnmpStatusException x) {
		    e = x;
		}
	    }
	    
	    synchronized(this) {
		if(counter >= 100) {
		    totalCounter += counter;
		    System.out.println("\tTotal received [" + 
				totalCounter +"] for [" + name + "]");
		    counter = 0;
		}
	    }
	}
    }
    
    private static class SnmpTable {
	private SnmpTableInfo description = null;
	private String name = null;
	private SnmpVarBindList columns = new SnmpVarBindList();
	private String oid = null;
	
	SnmpTable(String oid, String name, SnmpTableInfo description) {
	    this.oid = oid;
	    this.name = name;
	    this.description = description;
	}
	
	void addColumn(SnmpOidRecord record) {
	    try {
		columns.addVarBind(new SnmpVarBind(record.getOid()));
	    }catch(Exception e) {
		//What TODO
	    }
	}
	
	SnmpVarBindList getColumns() {
	    return columns;
	}
	
	SnmpTableInfo getDescription() {
	    if(description == null) {
		description = new SnmpTableInfo(new SnmpOid(oid),
						null,
						columns.size(),
						//-1,
						false,
						-1,
						null,
						true);
	    }
	    
	    return description;
	}

	String getName() {
	    return name;
	}
	
	String getOid() {
	    return oid;
	}
    }
    
    private static class SnmpTypeChecker {
	
	static HashMap checkers = new HashMap();
	
	static { 
	    checkers.put("C", new SnmpTypeChecker("C", SnmpCounter.class));
	    checkers.put("C64", new SnmpTypeChecker("C64", SnmpCounter64.class));
	    checkers.put("G", new SnmpTypeChecker("G", SnmpGauge.class));
	    checkers.put("I", new SnmpTypeChecker("I",SnmpInt.class));
	    checkers.put("IP", new SnmpTypeChecker("IP", SnmpIpAddress.class));
	    checkers.put("NU", new SnmpTypeChecker("NU", SnmpNull.class));
	    checkers.put("O", new SnmpTypeChecker("O", SnmpOpaque.class));
	    checkers.put("OI", new SnmpTypeChecker("OI", SnmpOid.class));
	    checkers.put("S", new SnmpTypeChecker("S", SnmpString.class));
	    checkers.put("T", new SnmpTypeChecker("T", SnmpTimeticks.class));
	    checkers.put("TA", new SnmpTypeChecker("TA", null));
	    checkers.put("EN", new SnmpTypeChecker("EN", null));
	    checkers.put("NT", new SnmpTypeChecker("NT", null));
	    checkers.put("ID", new SnmpTypeChecker("ID", null));
	    checkers.put("U", new SnmpTypeChecker("U", SnmpUnsignedInt.class));
	}
	
	private String type = null;
	private Class value = null;
	
	SnmpTypeChecker(String type, Class value) {
	    this.type = type;
	    this.value = value;
	}
	
	boolean isTable() {
	    return type.equals("TA");
	}
	
	Class getValueClass() {
	    return value;
	}
	
	boolean isValidSnmpValue(SnmpValue v) {
	    if(value == null)
		return false;
	    
	    if(v instanceof SnmpNull) {
		SnmpNull n = (SnmpNull) v;
		if(n.isNoSuchObjectValue()) {
		    logger.error("", "Received noSuchObject");
		    return false;
		}
		
		if(n.isNoSuchInstanceValue()) {
		    logger.error("", "Received noSuchInstance");
		    return false;
		}
	    }
	    return (value.isInstance(v));
	}
	
	boolean isGettableValue() {
	    /* "C" for Counter32
	     * "C64" for Counter64
	     * "G" for Gauge32
	     * "I" for Integer32
	     * "IP" for IpAddress
	     * "NU" for Null
	     * "O" for Opaque
	     * "OI" for Object Identifier
	     * "S" for String
	     * "T" for TimeTicks
	     * "U" for Unsigned32 
	     */
	    if(type.equals("C")   ||
	       type.equals("C64") ||
	       type.equals("G")   ||
	       type.equals("I")   ||
	       type.equals("IP")  ||
	       type.equals("NU")  ||
	       type.equals("O")   ||
	       type.equals("OI")  ||
	       type.equals("S")   ||
	       type.equals("T")   ||
	       type.equals("U"))
		return true;
	    
	    return false;
	}
    }
}

