/*
 * @(#)file      AsyncManager.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.23
 * @(#)lastedit  04/04/07
 *
 * Copyright 2004 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

// JDMK import
//
import com.sun.management.snmp.SnmpOid;
import com.sun.management.snmp.SnmpOidRecord;
import com.sun.management.snmp.SnmpVarBindList;
import com.sun.management.snmp.manager.SnmpPeer;
import com.sun.management.snmp.manager.SnmpParameters;
import com.sun.management.snmp.manager.SnmpRequest;
import com.sun.management.snmp.manager.SnmpSession;

import com.sun.management.snmp.SnmpEventReportDispatcher;
import com.sun.management.snmp.SnmpOidTableSupport;
import com.sun.jdmk.tasks.DaemonTaskServer;

/**
 * The class shows how to:
 * <br>     - use the OidTable generated by mibgen
 * <br>     - write and use an SnmpRequestHandler
 * <br>     - deal with asynchronous requests
 * <br>     - listen to traps sent by an agent
 */

public class AsyncManager {
   
    /**
     * Main entry point.
     * <p>When calling the program, you must specify:
     * <br> 1) host: hostname of the SNMP agent you want to query.
     * <br> 2) port: port number to use.
     *
     * <p> As a reminder, if you wish to query the SNMP agent example 
     *     provided as part of Java DMK, use port 8085.
     */
    public static void main(String argv[]) {
     
        if (argv.length != 2) {
            usage();
            java.lang.System.exit(1);
        }
    
        final String host = argv[0];
        final String port = argv[1];

        try {
       
            // Use the OidTable generated by mibgen when compiling MIB II.
            // This table makes it possible to translate symbolic names
	    // into OIDs.
            final SnmpOidTableSupport oidTable = new RFC1213_MIBOidTable();
       
            // Simple use of the OidTable.
            //
            final SnmpOidRecord record = 
		oidTable.resolveVarName("udpLocalPort");
            println("AsyncManager::main: variable = " + record.getName() + 
		    " oid = " + record.getOid() + 
		    " type = " + record.getType());
            
            // Initialize the SNMP Manager API.
            // Specify the OidTable containing all the MIB II knowledge.
            //
            SnmpOid.setSnmpOidTable(oidTable);
              
            // At any time, it is possible to add knowledge in the OidTable
            // using the SnmpOidDatabaseSupport.
            // For instance, we could start the Manager API like that:
            //    SnmpOidDatabaseSupport oidDB = 
	    //                new SnmpOidDatabaseSupport();
            //    SnmpOid.setSnmpOidTable(oidDB);
            // Then over time, MIB knowledge can be added as follows:
            //    oidDB.add(oidTable);
            //
            
            // Create an SnmpPeer object for representing the entity 
	    // to communicate with. 
            //
            final SnmpPeer agent = 
		new SnmpPeer(host, Integer.parseInt(port));
     
            // Create parameters to associate to the entity to 
	    // communicate with.
            // When creating the parameter object, you can specify the 
	    // read and write community to be used when querying the agent.
            // 
            final SnmpParameters params = 
		new SnmpParameters("public", "private");
       
            // The newly created parameter must be associated to the agent.
            //
            agent.setParams(params);
     
            // Build the session. A session creates, controls and manages
	    // one or more requests.
            //
            final SnmpSession session = 
		new SnmpSession("AsyncManager session");
       
            // A default peer (agent) can be associated to a SnmpSession. 
	    // When invoking a service provided by the SnmpSession, if 
	    // the agent is not specified, the session will perform
            // the service using the default peer as the target of the 
	    // service.
            //
            session.setDefaultPeer(agent);

	    // Create a taskServer for processing traps.
	    // This is an optional step. However using a DaemonTaskServer
	    // to process incomming PDUs makes it possible to empty
	    // the trap socket faster, thus reducing the hazards
	    // of trap loss.
	    // We set the priority of the DaemonTaskServer to 
	    // Thread.NORM_PRIORITY so that emptying the socket takes 
	    // precedence over trap processing.
	    //
	    final DaemonTaskServer taskServer = new DaemonTaskServer();
	    taskServer.start(Thread.NORM_PRIORITY);
	    
            // Create a listener and dispatcher for SNMP traps 
	    // (SnmpEventReportDispatcher).
            // SnmpEventReportDispatcher is run as a thread and listens 
	    // for traps in UDP port = agent port + 1.
            // Add TrapListenerImpl as SnmpTrapListener.
            // TrapListenerImpl will receive a callback when a valid 
	    // trap PDU is received.
            //
            final SnmpEventReportDispatcher trapAgent = 
		new SnmpEventReportDispatcher(Integer.parseInt(port)+1,
					      null,taskServer,null);
            trapAgent.addTrapListener(new TrapListenerImpl()); 
            final Thread trapThread = new Thread(trapAgent);
	    trapThread.setPriority(Thread.MAX_PRIORITY);
	    trapThread.start();

            // Build the list of variables you want to query.
            // For debug purposes, you can associate a name to your list.
            //
            final SnmpVarBindList list = 
		new SnmpVarBindList("AsyncManager varbind list");
     
            // We want to read the "sysDescr" variable.
	    //
            // We will thus query "sysDescr.0", as sysDescr is a scalar
	    // variable (see RFC 1157, section 3.2.6.3.  Identification 
            // of Object Instances, or RFC 2578, section 7.  Mapping of 
	    // the OBJECT-TYPE macro).
	    //
            list.addVarBind("sysDescr.0");
    
            // Create a simple implementation of a SnmpRequestHandler. 
            //
            final AsyncRspHandler handler = new AsyncRspHandler();
       
            // Make the SNMP walk request.
            // Read all the MIB variables starting at "sysDescr" and 
	    // ending up at "sysServices".
            // The responses will be received through the handler we 
	    // associated to the request.
            //
            final SnmpRequest request = 
		session.snmpWalkUntilRequest(handler, list, 
					     new  SnmpOid("sysServices"));
       
            println("\nAsyncManager::main: " + 
		    "Start snmpWalkUntilRequest for SNMP agent on " + 
		    host + " at port " + port);
       
            // Here you could do whatever processing you need, for 
	    // instance you could run a
            // graphical user interface.
            //
            Thread.sleep(4000);
       
            println("\n>> Press <Enter> if you want to stop" + 
		    " this SNMP manager.\n");
            java.lang.System.in.read();
            
            // Nicely stop the session
            //
            session.destroySession();
       
	    // Nicely stop the SnmpEventReportDispatcher.
	    //
	    trapAgent.close();
	    taskServer.terminate();

            // That's all !
            //
            java.lang.System.exit(0);
     
        } catch(Exception e) {
            java.lang.System.err.println("AsyncManager::main: "+ 
					 "Exception occurred:" + e );
            e.printStackTrace();
        }
    }
   
    /**
     * Return usage of the program.
     */
    public static void  usage() {
        println("java AsyncManager <host> <port>");
        println("where");
        println("\t-host: hostname or IP address of the SNMP agent" + 
		" you want to query.");
        println("\t-port: port number to use.");
    }

    /**
     * print/println stuff...
     */

    private final static void println(String msg) {
	java.lang.System.out.println(msg);
    }

    private final static void print(String msg) {
	java.lang.System.out.print(msg);
    }
}
