/*
 * @(#)file      Agent.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.32
 * @(#)lastedit  04/04/07
 *
 * Copyright 2004 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

// JMX imports
//
import javax.management.ObjectName;
import javax.management.MBeanServer;
import javax.management.MBeanServerFactory;

// JDMK imports
//
import com.sun.jdmk.comm.HtmlAdaptorServer;
import com.sun.management.comm.SnmpAdaptorServer;
import com.sun.management.snmp.agent.SnmpMibAgent;

/**
 * The Agent class provides an example on how to use the SNMP proxy.
 * This agent is able to serve local MIBs as well as remote MIBs 
 * (that is MIBs running outside the VM where the agent is running).
 * It acts as the single point of contacts for a set of MIBs or subagents.
 * 
 * A local DEMO_MIB is loaded and initialized.
 * An SNMP proxy is initialized and allows to query remote MIBs. 
 * 
 * The SNMP protocol adaptor of the agent is started on port 8085.
 * The SNMP proxy will query the specified host using the specified
 * port number. 
 */

public class Agent {

    /**
     * The main method allows you to specify the following optional
     * command-line parameters:
     *  1) host: hostname of the remote SNMP subagent you want to query.
     *  2) port: port number of the remote SNMP subagent you want to query.
     */
    public static void main(String argv[]) {

        MBeanServer server;
        ObjectName htmlObjName;
        ObjectName snmpObjName;
        ObjectName localMibObjName;
        ObjectName remoteMibObjName;
        String host;
        String port;
        int htmlPort = 8082;
        int snmpPort = 8085;

        if (argv.length != 2) {
            usage();
            java.lang.System.exit(1);
        }

        host = argv[0];
        port = argv[1];

        try {
            server = MBeanServerFactory.createMBeanServer();
            String domain = server.getDefaultDomain();

            // Create and start the HTML adaptor.
            //
            htmlObjName =
		new ObjectName(domain +
			       ":class=HtmlAdaptorServer,protocol=html,port=" +
			       htmlPort);
            java.lang.System.out.println("Adding HTML adaptor to MBean " +
					 "server with name \n\t" + htmlObjName);
            java.lang.System.out.println("NOTE: HTML adaptor is bound to " +
					 "TCP port " + htmlPort);
            HtmlAdaptorServer htmlAdaptor = new HtmlAdaptorServer(htmlPort);
            server.registerMBean(htmlAdaptor, htmlObjName);
            htmlAdaptor.start();

            // Create and start the SNMP adaptor.
            //
            snmpObjName =
		new ObjectName(domain +
			       ":class=SnmpAdaptorServer,protocol=snmp,port=" +
			       snmpPort);
            java.lang.System.out.println("Adding SNMP adaptor to MBean " +
					 "server with name \n\t" + snmpObjName);
            java.lang.System.out.println("NOTE: SNMP Adaptor is bound to " +
					 "UDP port " + snmpPort);
            SnmpAdaptorServer snmpAdaptor = new SnmpAdaptorServer(snmpPort);
            server.registerMBean(snmpAdaptor, snmpObjName);
            snmpAdaptor.start();

            // Create and initialize the local MIB Demo.
            //
            localMibObjName = new ObjectName("snmp:class=DEMO_MIB");
            java.lang.System.out.println("Adding DEMO_MIB to MBean server " +
					 "with name \n\t" + localMibObjName);
            DEMO_MIB localMib = new DEMO_MIB_IMPL();
            server.registerMBean(localMib, localMibObjName);
            java.lang.System.out.println("\nInitializing the MIB " +
					 localMibObjName);

            // Bind the SNMP adaptor to the MIB in order to make the MIB
	    // accessible through the SNMP protocol.
            //
            localMib.setSnmpAdaptor(snmpAdaptor);

            // Create and initialize the SNMP proxy.
            //
            remoteMibObjName = new ObjectName("snmp:class=proxy");
            java.lang.System.out.println("Adding proxy to MBean server " +
					 "with name \n\t" + remoteMibObjName);
            SnmpMibAgentImpl remoteMib = new SnmpMibAgentImpl();
            server.registerMBean(remoteMib, remoteMibObjName);
            remoteMib.initializeProxy(host, Integer.parseInt(port),
				      "1.3.6.1.2.1");
            java.lang.System.out.println("\nInitializing the SNMP proxy " +
					 remoteMibObjName +  " to query host " +
					 host + " using port " + port);

            // Bind the SNMP adaptor to the MIB in order to make the MIB
	    // accessible through the SNMP protocol.
            //
            remoteMib.setSnmpAdaptor(snmpAdaptor);
        } catch (Exception e) {
            java.lang.System.err.println("\nException occurred: " + e );
            e.printStackTrace();
            java.lang.System.exit(1);
        }
    }

    public static void  usage() {
        java.lang.System.out.println("\nUsage: java Agent <remote_host> " +
				     "<remote_port>");
        java.lang.System.out.println("where");
        java.lang.System.out.println("\t-remote_host: hostname of the remote " +
				     "SNMP subagent you want to query.");
        java.lang.System.out.println("\t              This remote agent can " +
				     "be either the StandAloneAgent or an " +
				     "SNMP daemon.");
        java.lang.System.out.println("\t-remote_port: port number of the " +
				     "remote SNMP subagent you want to query.");
    }
}
