/*
 * @(#)file      Agent.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.12
 * @(#)lastedit  04/04/07
 *
 * Copyright 2004 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

// JMX imports
//
import javax.management.ObjectName;
import javax.management.MBeanServer;
import javax.management.MBeanServerFactory;

// JDMK imports
//
import com.sun.management.snmp.SnmpPdu;
import com.sun.management.snmp.SnmpPduPacket;
import com.sun.management.snmp.SnmpStatusException;

import com.sun.jdmk.comm.HtmlAdaptorServer;
import com.sun.management.comm.SnmpAdaptorServer;

/**
 * The Agent class provides a simple example on how to use the SNMP
 * protocol adaptor.
 *
 * A subset of MIB II (RFC1213) is implemented. The MIB is loaded and
 * initialized.
 *
 * As such you can now see the MIB using your favorite SNMP manager,
 * or you can use a web browser and see the MIB through the HTML adaptor.
 *
 * In this example, the SNMP adaptor is started on port 8085,
 * and the traps are sent to the port 8086, i.e non standard ports for SNMP.
 * As such you do not need to be root to start the agent.
 */

public class Agent {

    static SnmpAdaptorServer snmpAdaptor = null;

    /**
     * This class is used to store request contextual user-data, such as
     * the community string supplied by the SNMP Manager in the SNMP
     * request PDU.
     *
     * You may extract as much information as you want from the request PDU,
     * but you must take care of synchronization problems since this object
     * can be potentially accessed by concurrent threads.
     *
     **/
    private static class UserData {

        /**
         * Public constructor
         *
         **/
        public UserData(SnmpPduPacket pdu) {

            // Extract the community string from the PDU,
            // and store it in this UserData object for later use.
            //
            if (pdu == null) return;
            if (pdu.community == null) return;
            community = new String(pdu.community);
        }

        /**
         * Returns the community string extracted from the original
         * SNMP request PDU.
         *
         * @returns The manager supplied community string.
         *
         **/
        public synchronized String getCommunity() {
            return community;
        }

        /**
         * The manager supplied community string extracted from the 
         * SNMP request PDU
         */
        protected String community = null;
    }

    /**
     * This class is a custom implementation of the 
     * {@link com.sun.management.snmp.agent.SnmpUserDataFactory} interface.
     *
     * It creates and returns contextual UserData objects that can be 
     * later accessed when processing the SNMP request.
     *
     * @see com.sun.management.snmp.agent.SnmpUserDataFactory
     **/
    private static class SnmpUserDataFactoryImpl 
        implements com.sun.management.snmp.agent.SnmpUserDataFactory {
        
        /**
         * Public constructor
         **/
        public SnmpUserDataFactoryImpl() {
        }

        /**
         * Allocates a new UserData object
         *
         * This method is called by the <CODE> SnmpAdaptorServer</CODE>
         * SNMP adaptor after an
         * incoming SnmpPduPacket has been decoded, and before 
         * dispatching the request to the MIB.
         *
         * This UserData object will be associated to the 
         * threads spawn by the SNMP Adaptor for this requests.
         *
         * @return a UserData object.
         **/ 
        public Object allocateUserData(SnmpPduPacket requestPdu)
            throws SnmpStatusException {
            return allocateUserData((SnmpPdu) requestPdu);
        }

        /**
         * This method is called by the <CODE> SnmpAdaptorServer</CODE>
         * SNMP Adaptor when the request
         * is completed so that allocated resources can be freed.
         *
         * Nothing to be done - we do not need to free any resource in
         * this implementation.
         *
         **/
        public void releaseUserData(Object userData, SnmpPduPacket responsePdu)
            throws SnmpStatusException {
        }

        /**
         * This method is called by the <CODE> SnmpV3AdaptorServer</CODE>
         * SNMP Adaptor when the request
         * Allocates a new UserData object
         *
         * This method is called by the SNMP adaptor after an
         * incoming SnmpPduPacket has been decoded, and before 
         * dispatching the request to the MIB.
         *
         * This UserData object will be associated to the 
         * threads spawn by the SNMP Adaptor for this requests.
         *
         * @return a UserData object.
         **/ 
        public Object allocateUserData(SnmpPdu requestPdu)
            throws SnmpStatusException {
            return new UserData((SnmpPduPacket) requestPdu);
        }

        /**
         * This method is called by the <CODE> SnmpV3AdaptorServer</CODE>
         * SNMP Adaptor when the request
         * is completed so that allocated resources can be freed.
         *
         * Nothing to be done - we do not need to free any resource in
         * this implementation.
         *
         **/
        public void releaseUserData(Object userData, SnmpPdu responsePdu)
            throws SnmpStatusException {
        }
    }

    /**
     * This methods shows how to retrieve the community string stored
     * in the UserData object associated with the current thread. Note
     * that this method will not work if it is called outside of an
     * SNMP Adaptor spawn thread.
     *
     * @return the manager supplied community string.
     **/
    public static String getManagerCommunity() {

        // Get the UserData associated with the current thread.
        //
        final Object userData = com.sun.jdmk.ThreadContext.get("SnmpUserData");

        if (userData == null) {

            // This may happen if we are not in a thread spawn by
            // the SnmpAdaptorServer (e.g.: we are not executing any
            // SNMP request in the current thread)
            //
            java.lang.System.out.println("No UserData available");
            return null;
        }

        if (!(userData instanceof UserData)) {

            // Should not happen - this would be a bug.
            //
            java.lang.System.out.println("Unexpected class for UserData");
            return null;
        }

        // returns the community string stored in the UserData
        //
        return ((UserData) userData).getCommunity();
    }

    /**
     * Main entry point.
     */
    public static void main(String args[]) {

        MBeanServer server;
        ObjectName htmlObjName;
        ObjectName snmpObjName;
        ObjectName mibObjName;
        ObjectName trapGeneratorObjName;
        int htmlPort = 8082;
        int snmpPort = 161;

        try {
            server = MBeanServerFactory.createMBeanServer();
            String domain = server.getDefaultDomain();

            // Create and start the HTML adaptor.
            //
            htmlObjName =
                new ObjectName(domain +
                               ":class=HtmlAdaptorServer,protocol=html,port=" +
                               htmlPort);
            java.lang.System.out.println(
                      "Adding HTML adaptor to MBean server with name \n\t" +
                      htmlObjName);
            java.lang.System.out.println("NOTE: HTML adaptor is bound to " +
                                         "TCP port " + htmlPort);
            HtmlAdaptorServer htmlAdaptor = new HtmlAdaptorServer(htmlPort);
            server.registerMBean(htmlAdaptor, htmlObjName);
            htmlAdaptor.start();

            //
            // SNMP specific code:
            //

            // Create and start the SNMP adaptor.
            // Specify the port to use in the constructor.
            // If you want to use the standard port (161)
            // comment out the line:
            //
            // snmpPort = 8085;
            //
            snmpPort = 8085;
            snmpObjName =
                new ObjectName(domain +
                               ":class=SnmpAdaptorServer,protocol=snmp,port=" +
                               snmpPort);
            java.lang.System.out.println(
                      "Adding SNMP adaptor to MBean server with name \n\t" +
                      snmpObjName);
            java.lang.System.out.println("NOTE: SNMP Adaptor is bound to " +
                                         "UDP port " + snmpPort);
            snmpAdaptor = new SnmpAdaptorServer(snmpPort);
            server.registerMBean(snmpAdaptor, snmpObjName);

            // Create an SnmpUserDataFactory and register it with
            // the SNMP Adaptor.
            //
            snmpAdaptor.setUserDataFactory(new SnmpUserDataFactoryImpl());

            // Start the SNMP Adaptor.
            //
            snmpAdaptor.start();

            // Create the MIB II (RFC 1213) and add it to the MBean server.
            //
            mibObjName= new ObjectName("snmp:class=RFC1213_MIB");
            java.lang.System.out.println(
                      "Adding RFC1213-MIB to MBean server with name \n\t" +
                      mibObjName);

            // Create an instance of the customized MIB
            //
            RFC1213_MIB mib2 = new RFC1213_MIB_IMPL();
            server.registerMBean(mib2, mibObjName);

            // Bind the SNMP adaptor to the MIB in order to make the MIB
            // accessible through the SNMP protocol adaptor.
            // If this step is not performed, the MIB will still live in
            // the Java DMK agent:
            // its objects will be addressable through HTML but not SNMP.
            //
	    mib2.setSnmpAdaptor(snmpAdaptor);

        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * Return a reference to the SNMP adaptor server.
     */
    static public SnmpAdaptorServer getSnmpAdaptor() {
        return snmpAdaptor;
    }

    /**
     * Return usage of the program.
     */
    public static void  usage() {
        java.lang.System.out.println("java Agent");
    }
}
