/*
 * @(#)file      SyncManager.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.10
 * @(#)lastedit  04/04/19
 *
 * Copyright 2004 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

// JDMK imports
//
import com.sun.management.snmp.SnmpDefinitions;
import com.sun.management.snmp.SnmpOid;
import com.sun.management.snmp.SnmpVarBindList;
import com.sun.management.snmp.SnmpVarBind;
import com.sun.management.snmp.SnmpString;
import com.sun.management.snmp.manager.SnmpPeer;
import com.sun.management.snmp.manager.SnmpParameters;
import com.sun.management.snmp.manager.SnmpRequest;
import com.sun.management.snmp.manager.SnmpSession;
import com.sun.management.snmp.manager.SnmpEventReportDispatcher;

import com.sun.management.snmp.SnmpOidTableSupport;

/**
 * The class is a simple example showing how to use the SNMP Manager API.
 *
 * When calling the program, you must specify:
 *      - host: hostname of the SNMP agent you want to query.
 *      - port: port number to use.
 *
 * As a reminder, if you wish to query the SNMP agent example provided
 * as part of Java DMK, use port 8085.
 */

public class SyncManager {
   
    /**
     * Main entry point.
     * When calling the program, you must specify:
     *  1) host: hostname of the SNMP agent you want to query.
     *  2) port: port number to use.
     *
     * As a reminder, if you wish to query the SNMP agent example provided
     * as part of Java DMK, use port 8085.
     */
    public static void main(String argv[]) {
     
        if (argv.length != 2) {
            usage();
            java.lang.System.exit(1);
        }
    
        String host = argv[0];
        String port = argv[1];

        try {
            
            // Initialize the SNMP Manager API.
            // Specify the OidTable containing all the MIB II knowledge.
            // Use the OidTable generated by mibgen when compiling MIB II.
            //
            SnmpOidTableSupport oidTable = new RFC1213_MIBOidTable();
            SnmpOid.setSnmpOidTable(oidTable);
       
            // At any time, it is possible to add knowledge in the OidTable
            // using the SnmpOidDatabaseSupport.
            // For instance, we could start the Manager API like that:
            //    SnmpOidDatabaseSupport oidDB = new SnmpOidDatabaseSupport();
            //    SnmpOid.setSnmpOidTable(oidDB);
            // Then over time, MIB knowledge can be added as follows:
            //    oidDB.add(oidTable);
            //
            
            // Create a SnmpPeer object for representing the entity to 
	    // communicate with. 
            //
            SnmpPeer agent = new SnmpPeer(host, Integer.parseInt(port));
     
            // Create parameters to associate to the entity to 
	    // communicate with.
            // When creating the parameter object, you can specify the read
	    // and write community to
            // be used when querying the agent.
            // 
            SnmpParameters params = new SnmpParameters("public", "private");
       
            // The newly created parameter must be associated to the agent.
            //
            agent.setParams(params);
     
            // Build the session. A session creates, controls and manages 
	    // one or more requests.
            //
            SnmpSession session= new SnmpSession("SyncManager session");
       
            // A default peer (agent) can be associated to a SnmpSession.
	    // When invoking a service provided by the SnmpSession, if 
	    // the agent is not specified, the session will perform
            // the service using the default peer as the target of the service.
            //
            session.setDefaultPeer(agent);
     

	    // Declare temporary local variables
	    //
	    SnmpRequest     request     = null;
	    SnmpVarBindList result      = null;
	    boolean         completed   = false;
	    int             errorStatus = 0;

            // Build the list of variables you want to query.
            // For debug purposes, you can associate a name to your list.
            //
            SnmpVarBindList getList = 
		new SnmpVarBindList("SyncManager GET varbind list");
     
            // We want to read the "sysUpTime", "sysDescr", "sysContact",
	    // and "sysLocation" variables.
            //
            getList.addVarBind("sysUpTime.0");
            getList.addVarBind("sysDescr.0");
            getList.addVarBind("sysContact.0");
            getList.addVarBind("sysLocation.0");
    
            // Make the SNMP get request and wait for the result.
            //
            request = session.snmpGetRequest(null, getList);
            java.lang.System.out.println("SyncManager::main:" + 
					 " Send GET request to SNMP agent" +
					 " on " + host + " at port " + port);
            completed = request.waitForCompletion(10000);
	    
            // Check for a timeout of the request.
            //
            if (completed == false) {
                java.lang.System.out.println("SyncManager::main:" + 
					     " GET Request timed out." + 
					     " Check reachability of agent");
	 
                // Print request.
                //
                java.lang.System.out.println("GET Request: " + 
					     request.toString());
                java.lang.System.exit(0);
            }
     
            // Now we have a response. Check if the response contains an error.
            //
            errorStatus = request.getErrorStatus();
            if (errorStatus != SnmpDefinitions.snmpRspNoError) {
                java.lang.System.out.println("Error status = " + 
				  SnmpRequest.snmpErrorToString(errorStatus));
                java.lang.System.out.println("Error index = " + 
					     request.getErrorIndex());
                java.lang.System.exit(0);
            }
       
            // Now we shall display the content of the result.
            //
            result = request.getResponseVarBindList();
            java.lang.System.out.println("GET Result: \n" + result);

            java.lang.System.out.println("\n>> Press <Enter> to continue...\n");
            java.lang.System.in.read();

            // Build the list of variables you want to set.
            //
	    SnmpVarBindList setList = 
		new SnmpVarBindList("SyncManager SET varbind list");

            // We want to write the "sysContact" and "sysLocation" variables.
            //
	    SnmpVarBind sysContact = 
		new SnmpVarBind(new SnmpOid("sysContact.0"),
				new SnmpString("contact@somewhere.com"));
	    SnmpVarBind sysLocation = 
		new SnmpVarBind(new SnmpOid("sysLocation.0"),
				new SnmpString("Grenoble, FRANCE"));

            setList.addVarBind(sysContact);
            setList.addVarBind(sysLocation);
    
            // Make the SNMP set request and wait for the result.
            //
            request = session.snmpSetRequest(null, setList);
            java.lang.System.out.println("SyncManager::main:" + 
					 " Send SET request to SNMP agent" +
					 " on " + host + " at port " + port);
            completed = request.waitForCompletion(10000);
	    
            // Check for a timeout of the request.
            //
            if (completed == false) {
                java.lang.System.out.println("SyncManager::main:" + 
					     " SET Request timed out." + 
					     " Check reachability of agent");
	 
                // Print request.
                //
                java.lang.System.out.println("SET Request: " +
					     request.toString());
                java.lang.System.exit(0);
            }
     
            // Now we have a response. Check if the response contains an error.
            //
            errorStatus = request.getErrorStatus();
            if (errorStatus != SnmpDefinitions.snmpRspNoError) {
                java.lang.System.out.println("Error status = " + 
				  SnmpRequest.snmpErrorToString(errorStatus));
                java.lang.System.out.println("Error index = " + 
					     request.getErrorIndex());
                java.lang.System.exit(0);
            }
       
            // Now we shall display the content of the result.
            //
            result = request.getResponseVarBindList();
            java.lang.System.out.println("SET Result: \n" + result);
      

            java.lang.System.out.println("\n>> Press <Enter> if you want to " +
					 "stop this SNMP manager.\n");
            java.lang.System.in.read();
            
            // Nicely stop the session
            //
            session.destroySession();
       
            //
            // That's all !
            //
            java.lang.System.exit(0);
     
        } catch (Exception e) {
            java.lang.System.err.println("SyncManager::main: Exception " +
					 "occurred:" + e );
            e.printStackTrace();
        }
    }
   
    /**
     * Return usage of the program.
     */
    public static void  usage() {
        java.lang.System.out.println("java SyncManager <host> <port>");
        java.lang.System.out.println("where");
        java.lang.System.out.println("\t-host: hostname or IP address of " +
				     "the SNMP agent you want to query.");
        java.lang.System.out.println("\t-port: port number to use.");
    }
}
